import os
import threading
from jinja2.environment import Environment
from jinja2.loaders import PrefixLoader, FileSystemLoader
from jinja2.exceptions import TemplateNotFound
from ..settings import template_dirs


t_local = threading.local()


def jinja2_environment(loader=None, **options):
    key = 'quarrel_env'
    e = getattr(t_local, key, None)
    if e:
        return e
    loader = loader or FileSystemLoader(template_dirs)
    e = Environment(loader=loader, **options)
    setattr(t_local, key, e)
    return e


def load_sql(using, filename_or_sql, context=None):
    template = filename_or_sql
    if not template.endswith('.sql'):
        template = f'{template}.sql'
    env = jinja2_environment()
    try:
        t = env.get_template(template)
    except (TemplateNotFound, OSError):
        try:
            template = os.path.join(using, 'sql', template)
            with open(template, 'r', encoding='utf-8') as f:
                content = f.read()
            t = env.from_string(content)
        except (FileNotFoundError, OSError):
            t = env.from_string(filename_or_sql)
    if context:
        sql = t.render(**context)
    else:
        sql = t.render()
    return sql



from concentric.managers import CachingConnectionManager as ccm


def get_connection(using, use_cache=True):
    """
    returns a tuple with the first value being a boolean
    that indicates whether the connection should be closed
    after use and the second value being the live connection
    """
    return not use_cache, ccm.get_instance().connect(using, use_cache)



from ..connections import get_connection
from ..raw import open_cursor


def query(
        using,
        filename_or_sql,
        params=None,
        context=None,
        use_cache=True,
        select=True):
    """
    Executes a query and returns the results from fetchall
    as a list of dicts. the keys of each dictionary will all be
    lowercased.

    :param str using: the concentric alias for this connection
    :param str filename_or_sql: the /path/to/filename to render as a template,
        or the sql query that we want to execute
    :param list params: the list of parameters to use, the placeholders
        in the string will be connection-specific, e.g., `:` style parameters
        for oracle and `?` style parameters for mysql and mssql
    :param dict context: the context that will be used when rendering
        the template
    :param bool use_cache: if True, we will try to use / reuse a cached connection
        otherwise, we will pull a fresh connection
    :param bool select: if the query that we are running expects results returned
    """
    close, conn = get_connection(using, use_cache)
    cursor = None
    try:
        cursor = open_cursor(
            using, conn, filename_or_sql, params, context, use_cache, select)
        if select:
            header = [ x[0].lower() for x in cursor.description ]
            data = [ dict(zip(header, x)) for x in cursor.fetchall() ]
            return data
    finally:
        cursor.close()
        if close and conn:
            conn.close()


def query_many(using, filename_or_sql, params=None, context=None, use_cache=True):
    """
    executes a query and then yields multiple results back.  This is useful
    for systems like sql server or vertica where a query can generate
    multiple result sets.
    """
    close, conn = get_connection(using)
    cursor = None
    try:
        cursor = open_cursor(
            using, conn, filename_or_sql,
            params, context, use_cache)
        done = False
        while not done:
            if cursor.description:
                header = [ x[0].lower().strip() for x in cursor.description ]
                data = []
                for x in cursor.fetchall():
                    data.append(dict(zip(header, x)))
                yield data
            done = not cursor.nextset()
    finally:
        if cursor:
            cursor.close()
        if close and conn:
            conn.close()





from .. import cooked
from ..templates import load_sql
from concentric.managers import Alchemist


def query(
        using,
        filename_or_sql,
        params=None,
        context=None,
        use_cache=True,
        select=True):
    """
    Executes a query and returns the results from fetchall
    as a dataframe

    :param str using: the concentric alias for this connection
    :param str filename_or_sql: the /path/to/filename to render as a template,
        or the sql query that we want to execute
    :param list params: the list of parameters to use, the placeholders
        in the string will be connection-specific, e.g., `:` style parameters
        for oracle and `?` style parameters for mysql and mssql
    :param dict context: the context that will be used when rendering
        the template
    :param bool use_cache: if True, we will try to use / reuse a cached connection
        otherwise, we will pull a fresh connection
    :param bool select: if the query that we are running expects results returned
    """
    from pandas import DataFrame
    return DataFrame(cooked.query(
        using, filename_or_sql, params, context,
        use_cache, select))


def query_alchemy(
        using,
        filename_or_sql,
        params=None,
        context=None, **kwargs):
    """
    Executes a query and returns the results from fetchall
    as a dataframe

    :param str using: the concentric alias for this connection
    :param str filename_or_sql: the /path/to/filename to render as a template,
        or the sql query that we want to execute
    :param list params: the list of parameters to use, the placeholders
        in the string will be connection-specific, e.g., `:` style parameters
        for oracle and `?` style parameters for mysql and mssql
    :param dict context: the context that will be used when rendering
        the template
    """
    import pandas
    sql = load_sql(using, filename_or_sql, context)
    with Alchemist.read_only(using) as session:
        return pandas.read_sql_query(sql, session.bind, params=params, **kwargs)




from concentric.managers import Alchemist
from ..templates import load_sql


def query(using, filename_or_sql, params=None, context=None, select=True):
    sql = load_sql(using, filename_or_sql, context)
    fn = Alchemist.read_only
    if not select:
        fn = Alchemist.transactional
    with fn(using) as session:
        result = session.execute(sql, params)
        if select:
            keys = [ x.lower().strip() for x in result.keys() ]
            return [ dict(zip(keys, x)) for x in result.fetchall() ]
        return result



from ..connections import get_connection
from ..templates import load_sql


def open_cursor(
        using,
        conn,
        filename_or_sql,
        params=None,
        context=None,
        use_cache=True,
        select=True):
    """
    Executes a query and returns the results from fetchall
    as a list of tuples (e.g., the raw result)

    :param str using: the concentric alias for this connection
    :param conn: an open connection to the database
    :param str filename_or_sql: the /path/to/filename to render as a template,
        or the sql query that we want to execute
    :param list params: the list of parameters to use, the placeholders
        in the string will be connection-specific, e.g., `:` style parameters
        for oracle and `?` style parameters for mysql and mssql
    :param dict context: the context that will be used when rendering
        the template
    :param bool use_cache: if True, we will try to use / reuse a cached connection
        otherwise, we will pull a fresh connection
    :param bool select: if the query that we are running expects results returned
    """
    sql = load_sql(using, filename_or_sql, context)
    cursor = conn.cursor()
    if params:
        cursor.execute(sql, params)
    else:
        cursor.execute(sql)
    return cursor


def query(
        using,
        filename_or_sql,
        params=None,
        context=None,
        use_cache=True,
        select=True):
    """
    Executes a query and returns the results from fetchall
    as a list of tuples (e.g., the raw result)

    :param str using: the concentric alias for this connection
    :param str filename_or_sql: the /path/to/filename to render as a template,
        or the sql query that we want to execute
    :param list params: the list of parameters to use, the placeholders
        in the string will be connection-specific, e.g., `:` style parameters
        for oracle and `?` style parameters for mysql and mssql
    :param dict context: the context that will be used when rendering
        the template
    :param bool use_cache: if True, we will try to use / reuse a cached connection
        otherwise, we will pull a fresh connection
    :param bool select: if the query that we are running expects results returned
    """
    close, conn = get_connection(using, use_cache)
    cursor = None
    try:
        cursor = open_cursor(
            using, conn, filename_or_sql, params, context, use_cache, select)
        if select:
            results = cursor.fetchall()
            return cursor.description, results
    finally:
        cursor.close()
        if close and conn:
            conn.close()


def query_many(using, filename_or_sql, params=None, context=None, use_cache=True):
    """
    executes a query and then yields multiple results back.  This is useful
    for systems like sql server or vertica where a query can generate
    multiple result sets.
    """
    close, conn = get_connection(using)
    cursor = None
    try:
        cursor = open_cursor(
            using, conn, filename_or_sql,
            params, context, use_cache)
        done = False
        while not done:
            if cursor.description:
                # header = [ x[0].lower().strip() for x in cursor.description ]
                # data = []
                # for x in cursor.fetchall():
                #     data.append(dict(zip(header, x)))
                yield cursor.description, list(cursor.fetchall())
            done = not cursor.nextset()
    finally:
        if cursor:
            cursor.close()
        if close and conn:
            conn.close()


